
#include <stdio.h>

typedef struct {
    float x;
    float y;
    float z;
} Point;

Point *splineCurve( n, prp_points, totalPoints )
int n;
Point *prp_points;
int *totalPoints;
{
   int i, j, frame;
   int inbetweens, nm1, nm2;
   int N[100];
   float x, y, z, delta;
   Point controlPoints[100];
   Point *splinePoints;
   float s[4], sH[4];

   static float Hermite[4][4] = { {  2., -2.,  1.,  1. },
                                  { -3.,  3., -2., -1. },
                                  {  0.,  0.,  1.,  0. },
                                  {  1.,  0.,  0.,  0. } };

   for( i=0, j=0; i<=n+1; i++, j++ ) {
      controlPoints[i].x = prp_points[j].x;
      controlPoints[i].y = prp_points[j].y;
      controlPoints[i].z = prp_points[j].z;
      N[i] = 15;
      if( !i ) i++;
   }

   n = i;
   nm1 = n - 1;
   nm2 = n - 2;

   controlPoints[1].x = controlPoints[0].x;
   controlPoints[1].y = controlPoints[0].y;
   controlPoints[1].z = controlPoints[0].z;
   N[1] = N[0];
   controlPoints[nm1].x = controlPoints[nm2].x;
   controlPoints[nm1].y = controlPoints[nm2].y;
   controlPoints[nm1].z = controlPoints[nm2].z;
   N[nm1] = N[nm2];

   for( i=0; i<n; i++ )
      *totalPoints += (N[i]+1);
   *totalPoints -= (N[1]+N[nm1]+N[nm1]+3);

   if( (splinePoints = (Point *)malloc( *totalPoints * sizeof(Point) )) ==
       (Point *)0 ) {
       fprintf( stderr, "Cant allocate Point\n" );
       exit(1);
   }

   j = 0;
   for( frame=1; frame<nm2; frame++ ) {
      inbetweens = N[frame];

      delta = 0.;
      s[0] = s[1] = s[2] = delta; s[3] = 1.;
      vector_mult( s, Hermite, sH );
      Spline( &splinePoints[j].x, &splinePoints[j].y, &splinePoints[j].z,
         frame, sH, N, controlPoints, 0., 0., 0. );
      j++;
      for( i=1; i<=inbetweens; i++ ) {

         delta = i/(float)(inbetweens);
         s[0] = delta * delta * delta; s[1] = delta * delta;
         s[2] = delta;                 s[3] = 1.;
         vector_mult( s, Hermite, sH );
         Spline( &splinePoints[j].x, &splinePoints[j].y, &splinePoints[j].z,
            frame, sH, N, controlPoints, 0., 0., 0. );
         j++;
      }
   }
   return splinePoints;
}


Spline( x, y, z, fr, sH, N, controlPoints, t, c, b )
float *x, *y, *z;
int fr;                                 /* at which frame we are at */
float sH[4];                            /* the s * Hermite matrix   */
int N[100];
Point controlPoints[100];
float t, c, b;
{
   float Pim1[3], Pi[3], Pi1[3], Pi2[3];
   float Di[3], Di1[3];
   float xvec[4], yvec[4], zvec[4];
   float xt, yt, zt;
   int i;
   int frm1, fr1, fr2;

   frm1 = fr-1;   fr1 = fr+1;   fr2 = fr+2;

   Pim1[0] = controlPoints[frm1].x;
   Pim1[1] = controlPoints[frm1].y;
   Pim1[2] = controlPoints[frm1].z;
   Pi  [0] = controlPoints[fr].x;
   Pi  [1] = controlPoints[fr].y;
   Pi  [2] = controlPoints[fr].z;
   Pi1 [0] = controlPoints[fr1].x;
   Pi1 [1] = controlPoints[fr1].y;
   Pi1 [2] = controlPoints[fr1].z;
   Pi2 [0] = controlPoints[fr2].x;
   Pi2 [1] = controlPoints[fr2].y;
   Pi2 [2] = controlPoints[fr2].z;

   for( i=0; i<3; i++ )
      Di[i]  = (( 0.5*(1-t)*(1+c)*(1+b) ) * ( Pi [i] - Pim1[i] ) +
                ( 0.5*(1-t)*(1-c)*(1-b) ) * ( Pi1[i] - Pi  [i] ))
                * (( 2. * N[frm1] ) / ( N[frm1] + N[fr] ));
   for( i=0; i<3; i++ )
      Di1[i] = (( 0.5*(1-t)*(1-c)*(1+b) ) * ( Pi1[i] - Pi [i] ) +
                ( 0.5*(1-t)*(1+c)*(1-b) ) * ( Pi2[i] - Pi1[i] ))
                * (( 2. * N[fr1] ) / ( N[fr] + N[fr1] ));

   xvec[0] = Pi [0];   xvec[1] = Pi1[0];   xvec[2] = Di [0];   xvec[3] = Di1[0];
   yvec[0] = Pi [1];   yvec[1] = Pi1[1];   yvec[2] = Di [1];   yvec[3] = Di1[1];
   zvec[0] = Pi [2];   zvec[1] = Pi1[2];   zvec[2] = Di [2];   zvec[3] = Di1[2];

   dotProduct( sH, xvec, &xt );
   dotProduct( sH, yvec, &yt );
   dotProduct( sH, zvec, &zt );
   *x = xt;   *y = yt;   *z = zt;
}


/*
 *  Procedure : get the cross product of two vectors
 */

dotProduct( u, v, res )
float u[4], v[4], *res;
{
   *res = u[0] * v[0] + u[1] * v[1] + u[2] * v[2] + u[3] * v[3];
}


/*
 *  Procedure : set equal two matrices
 */

matCopy( m1, m2 )
float *m1, *m2;
{
   int i;

   for ( i=0; i<16; i++ )
      *m2++ = *m1++;
}


/*
 *  Procedure : multiply a vector times a matrix
 *              result can be m1
 */

vector_mult( m1, m2, result )
float m1[4], m2[4][4], result[4];
{
   int i, j;
   float mt[4];

   for ( i=0; i<4; i++ ) {
      mt[i] = 0.;
      for ( j=0; j<4; j++ )
         mt[i] += m1[j] * m2[j][i];
   }
   vecCopy( mt, result );
}


/*
 *  Procedure : multiply a matrix times a matrix
 *              result can be m1
 */

matr_mult( m1, m2, result )
float m1[4][4], m2[4][4], result[4][4];
{
   int i, j, k;
   float mt[4][4];

   for ( i=0; i<4; i++ ) {
      for ( j=0; j<4; j++ ) {
         mt[i][j] = 0.;
         for ( k=0; k<4; k++ )
            mt[i][j] += m1[i][k] * m2[k][j];
      }
   }
   matCopy( mt, result );
}
